/*
 * PolyvalentLuceneLookup.java
 *
 * Created on 1 dcembre 2002, 11:57
 */

package fr.gouv.culture.sdx.lookup;

import fr.gouv.culture.sdx.lookup.*;
import fr.gouv.culture.sdx.document.*;
import fr.gouv.culture.sdx.exception.SDXException;
import fr.gouv.culture.sdx.exception.SDXExceptionCode;
import fr.gouv.culture.sdx.utils.database.DatabaseEntity;
import fr.gouv.culture.sdx.utils.database.LuceneDatabase;
import fr.gouv.culture.sdx.repository.Repository;
import fr.gouv.culture.sdx.utils.Utilities;

import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.Hits;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;

/**
 *
 * @author  pierrick
 */
public class PolyvalentLuceneLookup implements DocumentLookup, RelationLookup, BatchableLookup {

    private LuceneDatabase lookup;

    /** Avalon logger to write information. */
    protected org.apache.avalon.framework.logger.Logger logger;

    //Constants originally in LuceneDocumentBase

    /** A constant for the repository property in the database entities. */
    public final static String REPO_PROPERTY = "repo";
    /** A constant for the repository property in the database entities. */
    public final static String ORIGINAL_DOCUMENT_PROPERTY = "t_doc";
    /** A constant for the path property for attached documents. */
    public static final String MIMETYPE_PROPERTY = "mimetype";
    /** A constant for the path property for the type of document, Binary, HTML, or XML. */
    public final String DOCUMENT_TYPE = "sdxDocType";

   /** Property names for the relation index */ 
   private final String RELATION_PROPERTY_DOC1ID = "doc1id"; 
   private final String RELATION_PROPERTY_DOC1REPOSITORY = "doc1repository"; 
   private final String RELATION_PROPERTY_DOC2ID = "doc2id";   
   private final String RELATION_PROPERTY_DOC2REPOSITORY = "doc2repository"; 
   private final String RELATION_PROPERTY_TYPE= "type"; 
   private final String RELATION_PROPERTY_ORDER = "order"; 


    private boolean inBatch = false;
    private int batchCount = 0;
    private int batchSize = 50;
    
    public PolyvalentLuceneLookup(File directory) throws SDXException {        
        lookup = new LuceneDatabase(directory);        
        lookup.init();        
    }

    public void enableLogging(org.apache.avalon.framework.logger.Logger logger) {
        logger = logger;
    }

    public void addDocument(Document doc) throws SDXException {
        try {
            //creating a document entry in repository look-up index
            //TODO : take account of doc's repository -pb
            DatabaseEntity docLookupEntry = new DatabaseEntity(doc.getId());
            //TODO : check that no entry already exists for this id ? -pb
            docLookupEntry.enableLogging(this.logger);

            //saving the repository
            docLookupEntry.addProperty(REPO_PROPERTY, doc.getRepositoryForStorage());
            //saving the document type
            docLookupEntry.addProperty(DOCUMENT_TYPE, doc.getDocType());
            //saving the mime type
            if (Utilities.checkString(doc.getMimeType()))
                docLookupEntry.addProperty(MIMETYPE_PROPERTY, doc.getMimeType());
            else
                /*
                * TODO : does it have sense ?
                * it is assumed that every IndexableDocument has a mimetype
                * wouldn't it be consistent to also provide one for binary documents ? -pb
                */
                docLookupEntry.addProperty(MIMETYPE_PROPERTY, BinaryDocument.DEFAULT_MIMETYPE);

            //saving document entry in look-up index
            lookup.save(docLookupEntry);

            //Batch management
            if (inBatch) {
                batchCount++;
                if (batchCount >= getBatchSize()) doBatch();
            }

        } catch (SDXException e) {
            throw e;
        }
    }

    public void deleteDocument(Document doc) throws SDXException {
        try {
            //TODO : take account of doc's repository -pb
            DatabaseEntity docLookupEntry = lookup.getEntity(doc.getId());
            //TODO : check docLookupEntry validity -pb
            docLookupEntry.enableLogging(this.logger);
            lookup.delete(docLookupEntry);

            //Batch management
            if (inBatch) {
                batchCount++;
                if (batchCount >= getBatchSize()) doBatch();
            }

        } catch (SDXException e) {
            throw e;
        }
    }

    public boolean documentExists(Document doc) {
        try {
            //TODO : take account of doc's repository -pb
            DatabaseEntity docLookupEntry = lookup.getEntity(doc.getId());
            return (docLookupEntry != null);
        } catch (SDXException e) {
            return false;
        }        
    }

    public String getDocumentProperty(Document doc, String property) {
        try {
            //TODO : take account of doc's repository -pb
            DatabaseEntity docLookupEntry = lookup.getEntity(doc.getId());
            //TODO : check docLookupEntry validity -pb
            docLookupEntry.enableLogging(this.logger);
            return docLookupEntry.getProperty(property);
        } catch (SDXException e) {
            return null;
        }
    }

    public void addRelation(Document doc1, Document doc2, String relationType, Integer relationPosition) throws SDXException {
        try {
            //creating a document entry in repository look-up index
            DatabaseEntity relationEntry = new DatabaseEntity(getRelationID(doc1, doc2, relationType));
            //TODO : check that no entry already exists for this id ? we have to be *very* confident into getRelationID() -pb
            relationEntry.enableLogging(this.logger);
            relationEntry.addProperty(RELATION_PROPERTY_DOC1ID, doc1.getId());
            relationEntry.addProperty(RELATION_PROPERTY_DOC1REPOSITORY, doc1.getRepositoryForStorage());
            relationEntry.addProperty(RELATION_PROPERTY_DOC2ID, doc2.getId());
            relationEntry.addProperty(RELATION_PROPERTY_DOC2REPOSITORY, doc2.getRepositoryForStorage());
            if (Utilities.checkString(relationType))
                relationEntry.addProperty(RELATION_PROPERTY_TYPE, relationType); //TODO : use constants
            if (relationPosition != null)
                relationEntry.addProperty(RELATION_PROPERTY_ORDER, relationPosition.toString()); //TODO : use constants

            //saving relation entry in look-up index
            lookup.save(relationEntry);

            //Batch management
            if (inBatch) {
                batchCount++;
                if (batchCount >= getBatchSize()) doBatch();
            }

        } catch (SDXException e) {
            throw e;
        }
    }

    public void deleteRelation(Document doc1, Document doc2, String relationType) throws SDXException {
        try {
            //Make a query
            BooleanQuery bq = new org.apache.lucene.search.BooleanQuery();
            //all terms are mandatory (for now : see getRelatedDocuments for a more gnric implementation)
            //if (doc1 != null) {
                TermQuery tq1 = new TermQuery(new Term(RELATION_PROPERTY_DOC1ID, doc1.getId()));
                //TermQuery tq2 = new TermQuery(new Term(RELATION_PROPERTY_DOC1REPOSITORY, doc1.getRepositoryForStorage()));
                bq.add(tq1, true, false);
                //bq.add(tq2, true, false);
            //}
            //if (doc2 != null) {
                TermQuery tq3 = new TermQuery(new Term(RELATION_PROPERTY_DOC2ID, doc2.getId()));
                //TermQuery tq4 = new TermQuery(new Term(RELATION_PROPERTY_DOC2REPOSITORY, doc2.getRepositoryForStorage()));
                bq.add(tq3, true, false);
                //bq.add(tq4, true, false);
            //}
            //if (relationType != null) {
                TermQuery tq5 = new TermQuery(new Term(RELATION_PROPERTY_TYPE, relationType)); //TODO : use constants
                bq.add(tq5, true, false);
            //}
            //Relation order is not significant

            //retrieving the relation between the given documents
            Hits h = lookup.search(bq);

            if (h == null || h.length() == 0) {
                //TODO : determinate the required behaviour : exception, do nothing...
                logger.warn("no relation to delete");
            } 
            else if (h.length() == 1) {
                try {
                    DatabaseEntity relationEntry = lookup.getEntity(h.doc(0));
                    //TODO : check relationEntry validity -pb
                    relationEntry.enableLogging(logger);
                    lookup.delete(relationEntry);

                    //TODO : should we update the subsequent relationPosition ? -pb

                    //Batch management
                    if (inBatch) {
                        batchCount++;
                        if (batchCount >= getBatchSize()) doBatch();
                    }

                } catch (IOException e) {
                    //could not retrieve the document
                    String[] args = new String[2];
                    args[0] = lookup.getIndexPath();
                    args[1] = e.getMessage();
                    throw new SDXException(logger, SDXExceptionCode.ERROR_LUCENE_RETRIEVE_DOCUMENT, args, e);
                }
            }    
            else {
                //More than one relation : determine the required behaviour
                logger.warn("more than one relation to delete : " + h.length());
            }            
        } catch (SDXException e) {
            throw e;
        }
    }

    //TODO : define a proper format -pb
    private String getRelationID(Document doc1, Document doc2, String relationType){
        StringBuffer ID = new StringBuffer();
        ID.append(doc1.getRepositoryForStorage());
        ID.append("/");
        ID.append(doc1.getId());
        ID.append("->");
        ID.append(doc2.getRepositoryForStorage());
        ID.append("/");
        ID.append(doc2.getId());
        ID.append(":");
        ID.append(relationType);
        return ID.toString();
    }

    public Document[] getRelatedDocuments(Document doc, Boolean asOrigin, String relationType) throws SDXException {
        try {   
            BooleanQuery mainQuery = new org.apache.lucene.search.BooleanQuery();
            if (asOrigin != null) {
                Term t1, t2;
                if (asOrigin.booleanValue()) {                
                    t1 = new Term(RELATION_PROPERTY_DOC1ID, doc.getId());
                    //t2 = new Term(RELATION_PROPERTY_DOC1REPOSITORY, doc.getRepositoryForStorage());
                }
                else {
                    t1 = new Term(RELATION_PROPERTY_DOC2ID, doc.getId());
                    //t2 = new Term(RELATION_PROPERTY_DOC2REPOSITORY, doc.getRepositoryForStorage());
                }
                TermQuery tq1 = new TermQuery(t1);
                //TermQuery tq2 = new TermQuery(t2);                
                mainQuery.add(tq1, true, false);
                //mainQuery.add(tq2, true, false);
            }
            else {
                Term t1, t2, t3, t4;                
                t1 = new Term(RELATION_PROPERTY_DOC1ID, doc.getId());
                //t2 = new Term(RELATION_PROPERTY_DOC1REPOSITORY, doc.getRepositoryForStorage());
                t3 = new Term(RELATION_PROPERTY_DOC2ID, doc.getId());
                //t4 = new Term(RELATION_PROPERTY_DOC2REPOSITORY, doc.getRepositoryForStorage());
                TermQuery tq1 = new TermQuery(t1);
                //TermQuery tq2 = new TermQuery(t2);
                TermQuery tq3 = new TermQuery(t3);
                //TermQuery tq4 = new TermQuery(t4);
                BooleanQuery sub1 = new org.apache.lucene.search.BooleanQuery();                
                sub1.add(tq1, true, false);
                //sub1.add(tq2, true, false);
                BooleanQuery sub2 = new org.apache.lucene.search.BooleanQuery();
                sub2.add(tq3, true, false);
                ///sub2.add(tq4, true, false);
                mainQuery.add(sub1, false, false);
                mainQuery.add(sub2, false, false);
            }

            if (relationType != null) { //Null here means "every relatoin type"
                TermQuery tq = new TermQuery(new Term(RELATION_PROPERTY_TYPE, relationType)); //TODO : use constants
                mainQuery.add(tq, true, false);
            }
            //Relation order is not significant

            //retrieving the relation between the given documents
            Hits h = lookup.search(mainQuery);
            if (h == null || h.length() == 0) {
                return null;                
            } 
            else {
                Document relatedDocs[] = new Document[h.length()];

                for (int i = 0; i < h.length(); i++) {
                    DatabaseEntity lookupEntry = null;
                    try {
                        lookupEntry = lookup.getEntity(h.doc(i));
                        lookupEntry.enableLogging(logger);
                        //Creates a dummy document
                        Document relatedDoc = new BinaryDocument();
                        doc.enableLogging(logger);
                        if (asOrigin != null) {
                            if (asOrigin.booleanValue()) {                
                                relatedDoc.setId(lookupEntry.getProperty(RELATION_PROPERTY_DOC2ID));
                                //relatedDoc.setRepositoryForStorage(lookupEntry.getProperty(RELATION_PROPERTY_DOC2REPOSITORY));
                            }
                            else {
                                relatedDoc.setId(lookupEntry.getProperty(RELATION_PROPERTY_DOC1ID));
                                //relatedDoc.setRepositoryForStorage(lookupEntry.getProperty(RELATION_PROPERTY_DOC1REPOSITORY));
                            }
                        }
                        else {
                            //Our doc is the origin...
                            if (lookupEntry.getProperty(RELATION_PROPERTY_DOC1ID).equals(doc.getId()) 
                                //&& lookupEntry.getProperty(RELATION_PROPERTY_DOC1REPOSITORY).equals(doc.getRepositoryForStorage())
                            ) {
                                //consider the destination
                                relatedDoc.setId(lookupEntry.getProperty(RELATION_PROPERTY_DOC2ID));
                                //relatedDoc.setRepositoryForStorage(lookupEntry.getProperty(RELATION_PROPERTY_DOC2REPOSITORY));
                            }
                            //Our doc is the destination...
                            else if (lookupEntry.getProperty(RELATION_PROPERTY_DOC2ID).equals(doc.getId()) 
                                //&& lookupEntry.getProperty(RELATION_PROPERTY_DOC2REPOSITORY).equals(doc.getRepositoryForStorage())
                            ) {
                                //consider the origin
                                relatedDoc.setId(lookupEntry.getProperty(RELATION_PROPERTY_DOC1ID));
                                //relatedDoc.setRepositoryForStorage(lookupEntry.getProperty(RELATION_PROPERTY_DOC1REPOSITORY));
                            }
                            else {
                                logger.warn("Inconsistency in relation query results !");
                                //Throw exception
                            }
                        }
                        relatedDocs[i] = relatedDoc;                        
                    } catch (IOException e) {
                        //could not retrieve the document for the int "i" from the hits for attached documents
                        String[] args = new String[2];
                        args[0] = lookup.getIndexPath();
                        args[1] = e.getMessage();
                        throw new SDXException(logger, SDXExceptionCode.ERROR_LUCENE_RETRIEVE_DOCUMENT, args, e);
                    }
                }
                return relatedDocs;
            }
        } catch (SDXException e) {
            throw e;
        }
    }

    public int getBatchSize() {
        return batchSize;
    }

    public void setBatchSize(int size) throws SDXException {
        if (size < 1) {            
            String[] args = new String[1];
            args[0] = String.valueOf(size);
            //TODO : create exception code ERROR_INVALID_BATCH_SIZE
            throw new SDXException(logger, 9999, args, null);
        }
        if (inBatch) {
            //TODO : create exception code ERROR_BATCH_IN_PROCESS
            throw new SDXException(logger, 9999, null, null);
        }
        batchSize = size;
    }

    public void beginBatch()  throws SDXException {
        if (inBatch) {            
            //TODO : create exception code ERROR_BATCH_IN_PROCESS
            throw new SDXException(logger, 9999, null, null);
        }
        inBatch = true;
    }

    public void endBatch() throws SDXException {
        //Process last batch entries
        if (batchCount > 0) doBatch();
        inBatch = false;
    }

    public boolean isBatchInProcess() {
        return inBatch;
    }

    private synchronized void doBatch() throws SDXException {
        lookup.optimize();        
        batchCount = 0;
    }

}
